#! python3
# -*- encoding: utf-8 -*-

import os
import re
import random
import ast
from rtsf.p_applog import logger
from rtsf import p_exception,p_compat
from rtsf.p_common import FileSystemUtils,CommonUtils,ModuleUtils,FileUtils
from rtsf.p_compat import numeric_types,builtin_str

variable_regexp = r"\$([\w_]+)"
function_regexp = r"\$\{([\w_]+\([\$\w\.\-_ =,]*\))\}"
function_regexp_compile = re.compile(r"^([\w_]+)\(([\$\w\.\-_ =,]*)\)$")


def extract_variables(content):
    """ extract all variable names from content, which is in format $variable
    @param (str) content
    @return (list) variable name list

    e.g. 
    print(extract_variables("abc")); # => []
    print(extract_variables("$variable")); # => ["variable"]
    print(extract_variables("http://$url")); # => ['url']
    print(extract_variables("/blog/$postid")); # => ["postid"]
    print(extract_variables("/$var1/$var2")); # => ["var1", "var2"]
    
    """
    try:
        return re.findall(variable_regexp, content)
    except TypeError:
        return []


def extract_functions(content):
    """ extract all functions from string content, which are in format ${fun()}
    @param (str) content
    @return (list) functions list

    e.g. 
    print(extract_functions('${func(5)}')); # => ["func(5)"]
    print(extract_functions('${func(a=1, b=2)}')); # => ["func(a=1, b=2)"]
    print(extract_functions('${func(a,b,c)}')); # => ['func(a,b,c)']
    print(extract_functions('/api/1000?_t=${get_timestamp()}')); # => ["get_timestamp()"]
    print(extract_functions('/api/${add(1, 2)}')); # => ["add(1, 2)"]
    print(extract_functions("/api/${add(1, 2)}?_t=${get_timestamp()}")); # => ["add(1, 2)", "get_timestamp()"]
    """
    try:
        return re.findall(function_regexp, content)
    except TypeError:
        return []


def parse_string_value(str_value):
    """ parse string to number if possible
    e.g. "123" => 123
         "12.2" => 12.3
         "abc" => "abc"
         "$var" => "$var"
    """
    try:
        return ast.literal_eval(str_value)
    except ValueError:
        return str_value
    except SyntaxError:
        # e.g. $var, ${func}
        return str_value


def parse_function(content):
    """ parse function name and args from string content.
    @param (str) content
    @return (dict) function name and args

    e.g. 
    print(parse_function("func()")); # => {'kwargs': {}, 'args': [], 'func_name': 'func'}
    print(parse_function("func(5)")); # => {'kwargs': {}, 'args': [5], 'func_name': 'func'}
    print(parse_function("func(a=1, b=2)")); # => {'kwargs': {'a': 1, 'b': 2}, 'args': [], 'func_name': 'func'}
    print(parse_function('func(a,b,c)')); # => {'kwargs': {}, 'args': ['a', 'b', 'c'], 'func_name': 'func'}
    """
    matched = function_regexp_compile.match(content)
    if not matched:
        raise p_exception.FunctionNotFound("{} not found!".format(content))

    function_meta = {
        "func_name": matched.group(1),
        "args": [],
        "kwargs": {}
    }

    args_str = matched.group(2).replace(" ", "")
    if args_str == "":
        return function_meta

    args_list = args_str.split(',')
    for arg in args_list:
        if '=' in arg:
            key, value = arg.split('=')
            function_meta["kwargs"][key] = parse_string_value(value)
        else:
            function_meta["args"].append(parse_string_value(arg))

    return function_meta


def substitute_variables_with_mapping(content, mapping):
    """ substitute variables in content with mapping
    e.g.
    @params
        content = {
            'request': {
                'url': '/api/users/$uid',
                'headers': {'token': '$token'}
            }
        }
        mapping = {"$uid": 1000}
    @return
        {
            'request': {
                'url': '/api/users/1000',
                'headers': {'token': '$token'}
            }
        }
    """
    if isinstance(content, bool):
        return content

    if isinstance(content, (numeric_types, type)):
        return content

    if not content:
        return content

    if isinstance(content, (list, set, tuple)):
        return [
            substitute_variables_with_mapping(item, mapping)
            for item in content
        ]

    if isinstance(content, dict):
        substituted_data = {}
        for key, value in content.items():
            eval_key = substitute_variables_with_mapping(key, mapping)
            eval_value = substitute_variables_with_mapping(value, mapping)
            substituted_data[eval_key] = eval_value

        return substituted_data

    # content is in string format here
    logger.log_debug(u"Will substitute: {} with {}".format(content, mapping))
    for var, value in mapping.items():        
        logger.log_debug(u"\t: {} - {}".format(content, {var:value}))
        if content == var:
            # content is a variable
            content = value
        else:
            if not isinstance(value, str):
                value = builtin_str(value)
            
            if isinstance(content, builtin_str):
                content = content.replace(var, value)

    return content


def parse_project_data(data, testset_path=None):
    """ parse project data and generate cartesian product
    @param data: list type            
            e.g.
                [
                    {'csv': 'username_password.csv', 'by': 'Sequential'}, 
                    {'csv': 'devices.csv', 'by': 'Random'}
                ]
    @param testset_path: testset file path, used for locating csv file
    @return cartesian product in list
    """
    testcase_parser = TestCaseParser(file_path=testset_path)
    
    parsed_parameters_list = []
    for da in data:
        if isinstance(da, dict) and da.get("csv"):
            csv_list_of_dict_data = testcase_parser.get_csv_data(da.get('csv'), fetch_method = da.get("by",'Sequential'))
            parsed_parameters_list.append(csv_list_of_dict_data)
            
    return CommonUtils.gen_cartesian_product(*parsed_parameters_list)


class TestCaseParser(object):
#     def __init__(self, action_class_name, preference_action_file):        
#         self._functions, self._variables = {}, {}
#         self.file_path = preference_action_file
#         _Actions = ModuleUtils.get_imported_module(action_class_name)                    
#         self.bind_functions(ModuleUtils.get_callable_class_method_names(_Actions.WebHttp))
#         self._variables = _Actions.WebHttp.glob
    
    def __init__(self, variables={}, functions={}, file_path=None):
        self._functions, self._variables = {}, {}
        self.update_binded_variables(variables)
        self.bind_functions(functions)
        self.file_path = file_path
                        
    def update_binded_variables(self, variables):
        """ bind variables to current testcase parser
        @param variable -> dict
            e.g.
            {"ip": "127.0.0.1"}
        """
        self._variables = variables

    def bind_functions(self, functions):
        """ bind functions to current testcase parser
        @param functions -> dict
            e.g.
            {"test": <function test at 0x03508B30>}
        """
        self._functions = functions
        
    def get_bind_variable(self, variable_name):
        """
        @return: the value of variable_name
        """
        return self._get_bind_item("variable", variable_name)


    def get_bind_function(self, func_name):
        """
        @param func_name: function name
        @return: object of func_name
        """
        return self._get_bind_item("function", func_name)

    def get_csv_data(self, csv_file_name, fetch_method="Sequential"):
        """ get csv data
        @note:  first line should be define variable in csv file
        @param csv_file_name: csv file name
        @param fetch_method: Sequential or Random
        @return: list of dict
        """
        parameter_file_path = os.path.join(
            os.path.dirname(self.file_path),
            "{}".format(csv_file_name)
        )
        csv_content_list = FileUtils.load_file(parameter_file_path)

        if fetch_method.lower() == "random":
            random.shuffle(csv_content_list)

        return csv_content_list

    def _get_bind_item(self, item_type, item_name):

        if item_type == "function":
            if item_name in self._functions:
                return self._functions[item_name]
            else:
                # is not keyword function, continue to search
                pass
        elif item_type == "variable":
            if item_name in self._variables:
                return self._variables[item_name]
        else:
            raise p_exception.ParamsError("bind item should only be function or variable.")

        try:
            # preference functions
            assert self.file_path is not None
            return ModuleUtils.search_conf_item(self.file_path, item_type, item_name)
        except (AssertionError, p_exception.FunctionNotFound):
            raise p_exception.ParamsError(
                "{} is not defined in bind {}s!".format(item_name, item_type))

    def eval_content_with_bind_actions(self, content):
        """ parse content recursively, each variable and function in content will be evaluated.

        @param content =>  any data structure with ${func} or $variable

        """
        if content is None:
            return None

        if isinstance(content, (list, tuple)):
            return [self.eval_content_with_bind_actions(item) for item in content]

        if isinstance(content, dict):
            evaluated_data = {}
            for key, value in content.items():
                eval_key = self.eval_content_with_bind_actions(key)
                eval_value = self.eval_content_with_bind_actions(value)
                evaluated_data[eval_key] = eval_value

            return evaluated_data

        if isinstance(content, p_compat.basestring):

            # content is in string format here
            content = content.strip()

            # replace functions with evaluated value
            # Notice: _eval_content_functions must be called before _eval_content_variables
            content = self._eval_content_functions(content)

            # replace variables with binding value
            content = self._eval_content_variables(content)

        return content

    def _eval_content_functions(self, content):
        functions_list = extract_functions(content)
        for func_content in functions_list:
            function_meta = parse_function(func_content)
            func_name = function_meta['func_name']

            args = function_meta.get('args', [])
            kwargs = function_meta.get('kwargs', {})
            args = self.eval_content_with_bind_actions(args)
            kwargs = self.eval_content_with_bind_actions(kwargs)

            func = self.get_bind_function(func_name)
            eval_value = func(*args, **kwargs)

            func_content = "${" + func_content + "}"

            if func_content == content:

                logger.log_debug(u"eval functions result: {} -> {}".format(func_content, eval_value))

                # content is a variable
                content = eval_value
            else:
                tmp = content

                # content contains one or many variables
                content = content.replace(
                    func_content,
                    p_compat.str(eval_value), 1
                )

                logger.log_debug(u"eval functions result: {} -> {}".format(tmp, content))

        return content

    def _eval_content_variables(self, content):
        variables_list = extract_variables(content)

        for variable_name in variables_list:
            variable_value = self.get_bind_variable(variable_name)

            if "${}".format(variable_name) == content:

                logger.log_debug(u"eval variables result: ${} -> {}".format(variable_name, variable_value))

                # content is a variable
                content = variable_value
            else:
                tmp = content

                # content contains one or several variables
                content = content.replace("${}".format(variable_name),p_compat.str(variable_value), 1)

                logger.log_debug(u"eval variables result: {} -> {}".format(tmp, content))

        return content


class YamlCaseLoader(object):
    overall_def_dict = {
        "api": {},
        "suite": {}
    }
    testcases_cache_mapping = {}

    def translate(self):
        """ usage:
            m = YamlCaseLoader(r"D:\auto\buffer\test.yaml")
            for i in m.translate():print(i)
        :return iterator (case_name, execute_function)

        @note:  this method is useless
        """
        if not self.check():
            return

        for idx in range(len(self.testcases)):
            testing = self.testcases[idx]
            case_id = testing.get("testcaseid")
            case_name = FileSystemUtils.get_legal_filename("%s[%s]" %(case_id,p_compat.str(testing[self.__case_title_field])))

            # executer actions
            execute_actionss = []
            for field in self.__executer_seq_fields:
                steps_info = testing.get(field)
                for execute_function in steps_info:
                    if not execute_function:
                        continue
                    execute_actionss.append(execute_function)
            yield (case_name, execute_actionss, idx)

    def check(self):
        """ usage:
            print(YamlModel(r"D:\auto\buffer\test.yaml").check()    )
        :return Ture/False
        """
        result = True
        self.testcases,invalid_cases = self.getYamlCasesValue()
        if invalid_cases:
            print("Waring: Yaml need available fields:")
            for k,v in invalid_cases.items():
                print("\t%s -> %r" %(k, v))
            result = False
        elif not self.testcases:
            print('Warning: Invalid Yaml Test Model.')
            result = False

        return result

    @staticmethod
    def load_dependencies(path_or_yamlfile):
        """ load all api and suite definitions.
        @param path_or_yamlfile:  dir path or yamlfile path where have api folder and suite folder
        """
        if os.path.isdir(path_or_yamlfile):
            # cases path
            path = os.path.join(os.path.abspath(path_or_yamlfile), "dependencies")
        else:
            # case file path
            path = os.path.join(os.path.dirname(os.path.abspath(path_or_yamlfile)), "dependencies")

        api_def_folder = os.path.join(path, "api")
        suite_def_folder = os.path.join(path, "suite")

        # load api definitions
        for test_file in FileUtils.load_folder_files(api_def_folder):
            YamlCaseLoader.load_api_file(test_file)

        # load suite definitions
        for suite_file in FileUtils.load_folder_files(suite_def_folder):
            suite = YamlCaseLoader.load_file(suite_file)

            if "def" not in suite["project"]:
                raise p_exception.ParamsError("def missed in suite file: {}!".format(suite_file))

            call_func = suite["project"]["def"]
            function_meta = parse_function(call_func)
            suite["function_meta"] = function_meta
            YamlCaseLoader.overall_def_dict["suite"][function_meta["func_name"]] = suite

    @staticmethod
    def load_api_file(file_path):
        """ load api definition from file and store in overall_def_dict["api"]
            @param file_path: yaml file path
            @return: store in overall_def_dict["api"]
        """
        api_items = FileUtils.load_file(file_path)
        if not isinstance(api_items, list):
            raise p_exception.FileFormatError("API format error: {}".format(file_path))

        for api_item in api_items:
            if not isinstance(api_item, dict) or len(api_item) != 1:
                raise p_exception.FileFormatError("API format error: {}".format(file_path))

            key, api_dict = api_item.popitem()
            if key != "api" or not isinstance(api_dict, dict) or "def" not in api_dict:
                raise p_exception.FileFormatError("API format error: {}".format(file_path))

            api_def = api_dict.pop("def")
            function_meta = parse_function(api_def)
            func_name = function_meta["func_name"]

            if func_name in YamlCaseLoader.overall_def_dict["api"]:
                logger.log_warning("API definition duplicated: {}".format(func_name))

            api_dict["function_meta"] = function_meta
            YamlCaseLoader.overall_def_dict["api"][func_name] = api_dict

    @staticmethod
    def load_file(yaml_file):
        """ load yaml file
        :param yaml_file: yaml file path
        :return: testset
        """
        testset = {
            "file_path": yaml_file,
            "project": {},
            "cases": [],
        }

        if not os.path.isfile(yaml_file):
            raise p_exception.FileNotFoundError("Not found testcase file {}.".format(yaml_file))

        try:
            test_cases = FileUtils.load_file(yaml_file)
            logger.log_debug(u"Yaml raw dict: {}".format(test_cases))

            for item in test_cases:
                if not isinstance(item, dict) or len(item) != 1:
                    raise p_exception.FileFormatError("Testcase format error: {}".format(yaml_file))

                key, test_block = item.popitem()
                if not isinstance(test_block, dict):
                    raise p_exception.FileFormatError("Testcase format error: {}".format(yaml_file))

                if key == "project":
                    testset["project"].update(test_block)
                    testset["name"] = test_block.get("module", "Default Test Set")

                elif key == "case":
#                     case_id = test_block.pop("id","")
#                     if not case_id:
#                         raise p_exception.ModelFormatError("Some cases do not have 'case_id'.")
#                     if not re.search("^[\w-]+$",case_id):
#                         raise p_exception.ModelFormatError("Invalid case_id: {}".format(case_id))

                    name = test_block.get("name")
                    if not name:
                        raise p_exception.ModelFormatError("Some cases do not have 'name'.")

                    test_block["name"] = name
                    if "api" in test_block:
                        ref_call = test_block["api"]
                        def_block = YamlCaseLoader._get_block_by_name(ref_call, "api")
                        YamlCaseLoader._override_block(def_block, test_block)
                        logger.log_debug(u"merged api block: {}".format(test_block))
                        testset["cases"].append(test_block)

                    elif "suite" in test_block:
                        ref_call = test_block["suite"]
                        block = YamlCaseLoader._get_block_by_name(ref_call, "suite")
                        logger.log_debug(u"extend suite block: {}".format(block["cases"]))
                        testset["cases"].extend(block["cases"])

                    else:
                        testset["cases"].append(test_block)

                else:
                    logger.log_warning("Unexpected block key: '{0}' in '{1}', should only be ['project' or 'case']".format(key, yaml_file))

        except:
            logger.log_error(CommonUtils.get_exception_error())
        finally:
            return testset

    @staticmethod
    def load_files(path):
        """ load yaml testcases from file path
        :param path: path could be in several type
            - absolute/relative file path
            - absolute/relative folder path
            - list/set container with file(s) and/or folder(s)
        :return testcase sets list, each testset is corresponding to a file
            [
                testset_dict_1,
                testset_dict_2
            ]
        """
        if isinstance(path, (list, set)):
            testsets = []

            for file_path in set(path):
                if "dependencies" in file_path:
                    continue
                testset = YamlCaseLoader.load_files(file_path)
                if not testset:
                    continue
                testsets.extend(testset)

            return testsets

        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)

        if path in YamlCaseLoader.testcases_cache_mapping:
            return YamlCaseLoader.testcases_cache_mapping[path]

        if os.path.isdir(path):
            files_list = FileUtils.load_folder_files(path)
            testcases_list = YamlCaseLoader.load_files(files_list)

        elif os.path.isfile(path):
            try:
                testset = YamlCaseLoader.load_file(path)

                if testset["cases"]:
                    testcases_list = [testset]
                else:
                    testcases_list = []
            except p_exception.FileFormatError:
                testcases_list = []

        else:
            logger.log_error(u"file not found: {}".format(path))
            testcases_list = []

        YamlCaseLoader.testcases_cache_mapping[path] = testcases_list
        return testcases_list

    @staticmethod
    def _get_block_by_name(ref_call, ref_type):
        """ get test content by reference name
        :param ref_call: e.g. api_v1_Account_Login_POST($UserName, $Password)
        :param ref_type: "api" or "suite"
        """
        function_meta = parse_function(ref_call)
        func_name = function_meta["func_name"]
        call_args = function_meta["args"]
        block = YamlCaseLoader._get_test_definition(func_name, ref_type)
        def_args = block.get("function_meta").get("args", [])

        if len(call_args) != len(def_args):
            raise p_exception.ParamsError("call args mismatch defined args!")

        args_mapping = {}
        for index, item in enumerate(def_args):
            if call_args[index] == item:
                continue

            args_mapping[item] = call_args[index]
            logger.log_info(u"{0} define： {1}={2}".format(ref_type.capitalize(), item,call_args[index]))

        if args_mapping:
            block = substitute_variables_with_mapping(block, args_mapping)
            logger.log_info(u"Substitute variables with mapping finished.")

        return block

    @staticmethod
    def _get_test_definition(name, ref_type):
        """ get expected api or suite.
        :param name: api or suite name
        :param ref_type: "api" or "suite"
        :return expected api info if found, otherwise raise ApiNotFound exception
        """
        block = YamlCaseLoader.overall_def_dict.get(ref_type, {}).get(name)

        if not block:
            err_msg = "{} not found!".format(name)
            if ref_type == "api":
                raise p_exception.ApiNotFound(err_msg)
            else:
                # ref_type == "suite"
                raise p_exception.SuiteNotFound(err_msg)

        return block

    @staticmethod
    def _override_block(def_block, current_block):
        """ override def_block with current_block
            @note: def_block is not effect if current_block has value
        """
        
        merge_keys = ("pre_command", "post_command", "verify")        
        merge_keys_value = [(key, def_block.get(key, []), current_block.get(key, [])) for key in merge_keys]
        merge_name = current_block.get("name")
        
        current_block.update(def_block)
        for key, define, current in merge_keys_value:
            if not current:
                current_block[key] = define
            else:
                current_block[key] = current
        current_block['name'] = merge_name                                


def is_testset(data_structure):
    """ check if data_structure is a testset
    testset should always be in the following data structure:
        {
            "name": "desc1",
            "project": {},
            "cases": [testcase11, testcase12]
        }
    """
    if not isinstance(data_structure, dict):
        return False

    if "name" not in data_structure or "cases" not in data_structure:
        return False

    if not isinstance(data_structure["cases"], list):
        return False

    return True


def is_testsets(data_structure):
    """ check if data_structure is testset or testsets
    testsets should always be in the following data structure:
        testset_dict
        or
        [
            testset_dict_1,
            testset_dict_2
        ]
    """
    if not isinstance(data_structure, list):
        return is_testset(data_structure)

    for item in data_structure:
        if not is_testset(item):
            return False

    return True



#! python3
# -*- encoding: utf-8 -*-


import locale
import codecs
import glob
import zipfile
import importlib
import types
import imp
import sys
import os
import io
import time
import re
import subprocess
import json
import inspect
import hashlib
import socket
import random
import string
import itertools
import csv
import yaml

from rtsf import p_compat, p_exception
from collections import OrderedDict

ConfigParser = p_compat.ConfigParser

filesystemencoding = sys.getfilesystemencoding()
encoding = "utf-8"
if sys.platform == 'win32':
    # Part code of IOBinding
    try:
        encoding = locale.getdefaultlocale()[1]
        codecs.lookup(encoding)
    except LookupError:
        pass
else:
    print("Waring: not window system.encoding is not setted.")
encoding = encoding.lower()


# def set_sys_encode(code):
#     if p_compat.is_py2:
#         import sys;reload(sys)
#         getattr(sys, "setdefaultencoding")(code)
#     else:
#         import importlib,sys
#         importlib.reload(sys)

def init_project_env(subject='Automation', proj_path=None, sysencoding="utf-8", debug=False):
    """ Set the environment for pyrunner """

#     if sysencoding:
#         set_sys_encode(sysencoding)

    if not proj_path:
        try:
            executable_file_path = os.path.dirname(os.path.abspath(inspect.stack()[-1][1]))
        except:
            executable_file_path = os.path.dirname(sys.path[0])
        finally:
            proj_path = executable_file_path

    p = os.path.join(proj_path,subject)

    proj_conf = {
        "sys_coding": sysencoding,
        "debug": debug,
        "module_name": os.path.splitext(os.path.basename(subject))[0],
        "cfg_file": os.path.join(p,"config.ini"),
        "path": {"root": p,
                  "case": os.path.join(p,"testcase"),
                  "data": os.path.join(p,"data"),
                  "buffer": os.path.join(p,"buffer"),
                  "resource": os.path.join(p,"resource"),
                  "tools": os.path.join(p,"tools"),
                  "rst": os.path.join(p,"result"),
                  "rst_log": os.path.join(p,"result","testcase"),
                  "rst_shot": os.path.join(p,"result","screenshots"),
            },
        }

    [FileSystemUtils.mkdirs(v) for v in proj_conf["path"].values()]
    sys.path.append(p) if os.path.isdir(p) else ""
    return proj_conf


class IntelligentWaitUtils(object):

    @staticmethod
    def until_cmd(listcmd, end_expects=None, save2logfile=None, coding = encoding):
        """执行系统命令,并等待执行完
            @param listcmd: 执行的命令，列表格式
            @param end_expects: 命令执行结束，在输出的最后一行，正则搜素期望值，并设置 结果标志
            @param save2logfile:  设置执行过程，保存的日志
            @param coding: 设置输出编码
        Sample usage:
            cmd = ["ping","127.0.0.1","-n","1"]
            until_cmd(cmd)
            until_cmd(cmd, save2logfile= r"d:\auto\buffer\t.tmp")
            until_cmd(cmd, end_expects=u"平均 = 0ms", save2logfile= r"d:\auto\buffer\t.tmp")
        """
        if end_expects and not isinstance(end_expects, p_compat.str):
            raise Exception("invalide unicode string: '%s'" %end_expects)

        lines = []
        result = False
        subp = subprocess.Popen(listcmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while subp.poll()==None:
            next_line = subp.stdout.readline().decode(coding)
            if next_line:
                # print(next_line)
                lines.append(next_line)
                if end_expects and re.search(end_expects, next_line):
                    result = True
                else:
                    result = False
        subp.stdout.close()

        if subp.returncode:
            result = False
            lines.append("sub command error code: %s" % subp.returncode)

        if save2logfile:
            with open(save2logfile, 'a') as f:
                f.writelines(lines)

        return result

    @staticmethod
    def until(method, timeout=30, message=''):
        """Calls the method until the return value is not False."""
        end_time = time.time() + timeout
        while True:
            try:
                value = method()
                if value:
                    return value
            except:
                pass
            time.sleep(1)
            if time.time() > end_time:
                break
        raise Exception(message)

    @staticmethod
    def until_not(method, timeout=30, message=''):
        """Calls the method until the return value is False."""
        end_time = time.time() + timeout
        while True:
            try:
                value = method()
                if not value:
                    return value
            except:
                return True
            time.sleep(1)
            if time.time() > end_time:
                break
        raise Exception(message)

    @staticmethod
    def wait_for_connection(ip="localhost", port=4444, timeout=30):
        sk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        end_time = time.time() + timeout
        while True:
            try:
                sk.connect((ip,int(port)))
                sk.shutdown(2)
                sk.close()
                return True
            except:
                pass
            time.sleep(1)
            if time.time() > end_time:
                break
        return False


class DateTimeUtils(object):

    @staticmethod
    def get_stamp_date():
        """ Return the current date """
        return time.strftime("%Y-%m-%d")

    @staticmethod
    def get_stamp_datetime():
        """ Return the current date time """
        return time.strftime("%Y-%m-%d %H:%M:%S")

    @staticmethod
    def get_stamp_datetime_coherent():
        """ Return the current date time """
        return time.strftime("%Y-%m-%d_%H_%M_%S")


class FileUtils(object):

    @staticmethod
    def _check_format(file_path, content):
        """ check testcase format if valid
        """
        if not content:
            # testcase file content is empty
            err_msg = u"Testcase file content is empty: {}".format(file_path)
            raise p_exception.FileFormatError(err_msg)

        elif not isinstance(content, (list, dict)):
            # testcase file content does not match testcase format
            err_msg = u"Testcase file content format invalid: {}".format(file_path)
            raise p_exception.FileFormatError(err_msg)

    @staticmethod
    def _load_yaml_file(yaml_file):
        """ load yaml file and check file content format
        """
        with io.open(yaml_file, 'r', encoding='utf-8') as stream:
            yaml_content = yaml.safe_load(stream)
            FileUtils._check_format(yaml_file, yaml_content)
            return yaml_content

    @staticmethod
    def _dump_yaml_file(data, yaml_file):
        """ dump data to yaml file
        """
        # FileUtils._check_format(yaml_file, data)
        with io.open(yaml_file, 'w', encoding='utf-8') as stream:
            yaml.dump(data, stream, allow_unicode=True)

    @staticmethod
    def _load_json_file(json_file):
        """ load json file and check file content format
        """
        with io.open(json_file, encoding='utf-8') as data_file:
            try:
                json_content = json.load(data_file)
            except p_exception.JSONDecodeError:
                err_msg = u"JSONDecodeError: JSON file format error: {}".format(json_file)
                raise p_exception.FileFormatError(err_msg)

            FileUtils._check_format(json_file, json_content)
            return json_content

    @staticmethod
    def _load_csv_file(csv_file):
        """ load csv file and check file content format
        @param
            csv_file: csv file path
            e.g. csv file content:
                username,password
                test1,111111
                test2,222222
                test3,333333
        @return
            list of parameter, each parameter is in dict format
            e.g.
            [
                {'username': 'test1', 'password': '111111'},
                {'username': 'test2', 'password': '222222'},
                {'username': 'test3', 'password': '333333'}
            ]
        """
        csv_content_list = []

        with io.open(csv_file, encoding='utf-8') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                csv_content_list.append(row)

        return csv_content_list

    @staticmethod
    def load_file(file_path):
        if not os.path.isfile(file_path):
            raise p_exception.FileNotFoundError("{} does not exist.".format(file_path))

        file_suffix = os.path.splitext(file_path)[1].lower()
        if file_suffix == '.json':
            return FileUtils._load_json_file(file_path)
        elif file_suffix in ['.yaml', '.yml']:
            return FileUtils._load_yaml_file(file_path)
        elif file_suffix == ".csv":
            return FileUtils._load_csv_file(file_path)
        else:
            # '' or other suffix
            print(u"Unsupported file format: {}".format(file_path))
            return []

    @staticmethod
    def load_folder_files(folder_path, recursive=True, end_with=None):
        """ load folder path, return all files in list format.
        @param
            folder_path: specified folder path to load
            recursive: if True, will load files recursively
            end_with: file extension, default is ('.yml', '.yaml', '.json')
        """
        end_with = ('.yml', '.yaml', '.json') if end_with is None else end_with

        if isinstance(folder_path, (list, set)):
            files = []
            for path in set(folder_path):
                files.extend(FileUtils.load_folder_files(path, recursive))

            return files

        if not os.path.exists(folder_path):
            return []

        file_list = []

        for dirpath, dirnames, filenames in os.walk(folder_path):
            filenames_list = []

            for filename in filenames:
                if not filename.endswith(end_with):
                    continue

                filenames_list.append(filename)

            for filename in filenames_list:
                file_path = os.path.join(dirpath, filename)
                file_list.append(file_path)

            if not recursive:
                break

        return file_list


class FileSystemUtils(object):

    @staticmethod
    def mkdirs(dir_path):
        """ make a directory if it not exists"""
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

    @staticmethod
    def get_file_md5(filePath):
        if not os.path.isfile(filePath):
            return
        myhash=hashlib.md5()
        with open(filePath,'rb') as f:
            while True:
                b=f.read(8096)
                if not b:
                    break;
                myhash.update(b)

        return myhash.hexdigest()

    @staticmethod
    def get_file_size(filePath):
        if not os.path.isfile(filePath):
            return
        else:
            return os.path.getsize(filePath)

    @staticmethod
    def get_legal_filename(fn):
        """
        @param fn: file name
        @return: legal file name
        """
        prog = re.compile(r"[\n\\/:*?\"<>|]")
        return prog.sub("", fn)

    @staticmethod
    def add_unique_postfix(fn):
        """
        @param fn: File name
        @return:  Return an unique postfix for the file name
        """

        fn = p_compat.str(fn)

        if not os.path.exists(fn):
            return fn

        path, name = os.path.split(fn)
        name, ext = os.path.splitext(name)

        make_fn = lambda i: os.path.join(path, '%s_%d%s' % (name, i, ext))

        for i in p_compat.xrange(2, sys.maxsize):
            uni_fn = make_fn(i)
            if not os.path.exists(uni_fn):
                return uni_fn

        return None

    @staticmethod
    def force_delete_file(file_path):
        """ force delete a file """
        if os.path.isfile(file_path):
            try:
                os.remove(file_path)
                return file_path
            except:
                return FileSystemUtils.add_unique_postfix(file_path)
        else:
            return file_path


class ZipUtils(object):

    @staticmethod
    def mkzip(source_dir, output_filename):
        """Usage:
            p = r'D:\auto\env\ttest\ins\build\lib\rock4\softtest\support'
            mkzip(os.path.join(p, "appiumroot"),os.path.join(p, "appiumroot.zip"))
            unzip(os.path.join(p, "appiumroot.zip"),os.path.join(p, "appiumroot2"))
        """
        zipf = zipfile.ZipFile(output_filename, 'w', zipfile.zlib.DEFLATED)
        pre_len = len(os.path.dirname(source_dir))
        for parent, dirnames, filenames in os.walk(source_dir):
            for filename in filenames:
                pathfile = os.path.join(parent, filename)
                arcname = pathfile[pre_len:].strip(os.path.sep);#相对路径
                zipf.write(pathfile, arcname)
        zipf.close()

    @staticmethod
    def unzip(zipfilename, unziptodir):
        if not os.path.exists(unziptodir): os.mkdir(unziptodir)

        zfobj = zipfile.ZipFile(zipfilename)

        for z_info in zfobj.infolist():
            z_name = z_info.filename.replace('\\', '/')

            if z_info.flag_bits == 2048:
                # For UTF8, such as ZipUtils.mkzip
                ext_filename = os.path.join(unziptodir, z_name)
                # ext_filename = os.path.join(unziptodir, name.encode('utf-8'))
                # print("UTF8:------", ext_filename)
            else:
                # For GBK
                ext_filename = os.path.join(unziptodir, z_name.encode('cp437').decode('gbk'))
                # print("GBK:------", ext_filename)

            if z_name.endswith(os.sep):
                os.mkdir(ext_filename)
            else:
                ext_dir, ext_file = os.path.split(ext_filename)
                if not os.path.exists(ext_dir):
                    os.makedirs(ext_dir)

                if ext_file == '':
                    continue

                with open(ext_filename, 'wb') as f:
                    f.write(zfobj.read(z_name))


class ModuleUtils(object):

    @staticmethod
    def get_callable_class_method_names(testClass):
        """
        @param testClass is a Class Object
        @return e.g.
            {"test": <function test at 0x03508B30>}
        """
        is_instance_method = lambda attrname: not attrname.startswith("_") and hasattr(getattr(testClass, attrname), '__call__')
        methods = filter(is_instance_method, dir(testClass))
        return {i: getattr(testClass, i) for i in methods}

    @staticmethod
    def is_function(tup):
        """ Takes (name, object) tuple, returns True if it is a function.
        """
        _, item = tup
        return isinstance(item, types.FunctionType)

    @staticmethod
    def is_variable(tup):
        """ Takes (name, object) tuple, returns True if it is a variable.
        """
        name, item = tup
        if callable(item):
            # function or class
            return False

        if isinstance(item, types.ModuleType):
            # imported module
            return False

        if name.startswith("_"):
            # private property
            return False

        return True

    @staticmethod
    def get_imported_module(module_name):
        """ import module and return imported module
        """
        return importlib.import_module(module_name)

    @staticmethod
    def get_imported_module_from_file(file_path):
        """ import module from python file path and return imported module
        """
        if p_compat.is_py3:
            imported_module = importlib.machinery.SourceFileLoader('module_name', file_path).load_module()
        elif p_compat.is_py2:
            imported_module = imp.load_source('module_name', file_path)
        else:
            raise RuntimeError("Neither Python 3 nor Python 2.")

        return imported_module

    @staticmethod
    def filter_module(module, filter_type):
        """ filter functions or variables from import module
        @params
            module: imported module
            filter_type: "function" or "variable"
        """
        filter_type = ModuleUtils.is_function if filter_type == "function" else ModuleUtils.is_variable
        module_functions_dict = dict(filter(filter_type, vars(module).items()))
        return module_functions_dict

    @staticmethod
    def search_conf_item(start_path, item_type, item_name):
        """ search expected function or variable recursive upward
        @param
            start_path: search start path
            item_type: "function" or "variable"
            item_name: function name or variable name
        e.g.
            search_conf_item('C:/Users/RockFeng/Desktop/s/preference.py','function','test_func')
        """
        dir_path = os.path.dirname(os.path.abspath(start_path))
        target_file = os.path.join(dir_path, "preference.py")

        if os.path.isfile(target_file):
            imported_module = ModuleUtils.get_imported_module_from_file(target_file)
            items_dict = ModuleUtils.filter_module(imported_module, item_type)
            if item_name in items_dict:
                return items_dict[item_name]
            else:
                return ModuleUtils.search_conf_item(dir_path, item_type, item_name)

        if dir_path == start_path:
            # system root path
            err_msg = "'{}' not found in recursive upward path!".format(item_name)
            if item_type == "function":
                raise p_exception.FunctionNotFound(err_msg)
            else:
                raise p_exception.VariableNotFound(err_msg)

        return ModuleUtils.search_conf_item(dir_path, item_type, item_name)


class SetupUtils(object):

    @staticmethod
    def find_data_files(source, target, patterns, isiter=False):
        """Locates the specified data-files and returns the matches;
            filesystem tree for setup's data_files in setup.py
            Usage:
                data_files = find_data_files(r"C:\Python27\Lib\site-packages\numpy\core","numpy/core",["*.dll","*.pyd"])
                data_files = find_data_files(r"d:\auto\buffer\test\test","buffer/test/test",["*"],True)
            :param source -a full path directory which you want to find data from
            :param target -a relative path directory which you want to pack data to
            :param patterns -glob patterns, such as "*dll", "*pyd"  etc.
            :param isiter - True/Fase, Will traverse path if True when patterns equal ["*"]
        """
        if glob.has_magic(source) or glob.has_magic(target):
            raise ValueError("Magic not allowed in src, target")
        ret = {}
        for pattern in patterns:
            pattern = os.path.join(source,pattern)
            for filename in glob.glob(pattern):
                if os.path.isfile(filename):
                    targetpath = os.path.join(target,os.path.relpath(filename, source))
                    path = os.path.dirname(targetpath)
                    ret.setdefault(path,[]).append(filename)
                elif isiter and os.path.isdir(filename):
                    source2 = os.path.join(source, filename)
                    targetpath2 = "%s/%s" % (target, os.path.basename(filename))
                    # iter_target = os.path.dirname(targetpath2)
                    ret.update(SetupUtils.find_data_files(source2, targetpath2, patterns, isiter))

        return sorted(ret.items())


class CommonUtils(object):

    @staticmethod
    def gen_random_string(str_len):
        return ''.join(
            random.choice(string.ascii_letters + string.digits) for _ in range(str_len))

    @staticmethod
    def gen_cartesian_product(*args):
        """ generate cartesian product for lists,  笛卡尔积
        @param
            (list) args
                [{"a": 1}, {"a": 2}],
                [
                    {"x": 111, "y": 112},
                    {"x": 121, "y": 122}
                ]
        @return
            cartesian product in list
            [
                {'a': 1, 'x': 111, 'y': 112},
                {'a': 1, 'x': 121, 'y': 122},
                {'a': 2, 'x': 111, 'y': 112},
                {'a': 2, 'x': 121, 'y': 122}
            ]
        """
        if not args:
            return []
        elif len(args) == 1:
            return args[0]

        product_list = []
        for product_item_tuple in itertools.product(*args):
            product_item_dict = {}
            for item in product_item_tuple:
                product_item_dict.update(item)

            product_list.append(product_item_dict)

        return product_list

    @staticmethod
    def convert_to_order_dict(map_list):
        """ convert mapping in list to ordered dict
        @param (list) map_list
            [
                {"a": 1},
                {"b": 2}
            ]
        @return (OrderDict)
            OrderDict({
                "a": 1,
                "b": 2
            })
        """
        ordered_dict = OrderedDict()
        for map_dict in map_list:
            ordered_dict.update(map_dict)

        return ordered_dict

    @staticmethod
    def get_value_from_cfg(cfg_file):
        """ initial the configuration with file that you specify
            Sample usage:
                config = get_value_from_cfg()
            return:
                return a dict        -->config[section][option]  such as config["twsm"]["dut_ip"]
        """

        if not os.path.isfile(cfg_file):
            return

        cfg = {}
        config = ConfigParser.RawConfigParser()

        try:
            config.read(cfg_file)
        except Exception as e:
            # raise Exception("\n\tcommon exception 1.2: Not a well format configuration file. error: '%s'" %(e))
            return
        for section in config.sections():
            cfg[section] = {}
            for option in config.options(section):
                cfg[section][option]=config.get(section,option)
        return cfg

    @staticmethod
    def get_sorted_list(ll):
        """ 按数字排序
        Sample usage:
            get_sort_list(["t1","t11","t2","t22","t3ss","t4gg"])
        """

        if not isinstance(ll, list):
            return ll

        re_digits = re.compile(r'(\d+)')

        def emb_numbers(s):
            pieces = re_digits.split(s)
            pieces[1::2] = map(int, pieces[1::2])
            return pieces

        return sorted(ll, key=emb_numbers)

    @staticmethod
    def get_exception_error():
        """ Get the exception info
        Sample usage:
            try:
                raise Exception("asdfsdfsdf")
            except:
                print common.get_exception_error()
        Return:
            return the exception infomation.
        """
        error_message = ""
        for i in range(len(inspect.trace())):
            error_line = u"""
        File:      %s - [%s]
        Function:  %s
        Statement: %s
        -""" % (inspect.trace()[i][1], inspect.trace()[i][2], inspect.trace()[i][3], inspect.trace()[i][4])

            error_message = "%s%s" % (error_message, error_line)

        error_message = u"""Error!\n%s\n\t%s\n\t%s\n""" % (error_message, sys.exc_info()[0], sys.exc_info()[1])
        error_message = error_message + "-" * 90 + "\n\n"
        return error_message


class ProgressBarUtils(object):

    @staticmethod
    def echo(transferred, toBeTransferred, suffix=''):
        """ usage:
            for i in range(101):
                ProgressBarUtils.echo(i,100)
        """
        bar_len = 60
        rate = transferred/float(toBeTransferred)

        filled_len = int(round(bar_len * rate))
        _percents = "%s%s" %(round(100.0 * rate, 1), "%")

        end_str = "\r"
        _bar = '=' * filled_len + '-' * (bar_len - filled_len)
        print("[%s] %s ...%s%s" %(_bar, _percents, suffix, end_str))

    def __init__(self, title, transferred=0.0, run_status=None, fin_status=None, toBeTransferred=100.0, unit='', sep='/', chunk_size=1.0):
        self.title = title
        self.toBeTransferred = toBeTransferred
        self.transferred = transferred
        self.chunk_size = chunk_size
        self.status = run_status or ""
        self.fin_status = fin_status or " " * len(self.statue)
        self.unit = unit
        self.seq = sep

    def echo_size(self, transferred=1, status=None):
        """Sample usage:

            f=lambda x,y:x+y
            ldata = range(10)
            toBeTransferred = reduce(f,range(10))

            progress = ProgressBarUtils("refresh", toBeTransferred=toBeTransferred, unit="KB", chunk_size=1.0, run_status="正在下载", fin_status="下载完成")
            import time
            for  i in ldata:
                time.sleep(0.2)
                progress.echo_size(transferred=i)
        """
        self.transferred += transferred
        # if status is not None:
        self.status = status or self.status
        end_str = "\r"
        if self.transferred == self.toBeTransferred:
            end_str = '\n'
            self.status = status or self.fin_status

        print(self.__get_info() + end_str)

    def echo_percent(self,transferred=1, status=None):
        """Sample usage:
            f=lambda x,y:x+y
            ldata = range(10)
            toBeTransferred = reduce(f,range(10))

            import time
            progress = ProgressBarUtils("viewbar", toBeTransferred=toBeTransferred, run_status="正在下载", fin_status="下载完成")
            for i in ldata:
                time.sleep(0.1)
                progress.echo_percent(i)
        """
        self.transferred += transferred
        self.status = status or self.status
        end_str = "\r"
        if self.transferred == self.toBeTransferred:
            end_str = '\n'
            self.status = status or self.fin_status
        print(self.__get_bar() + end_str)

    def __get_info(self):
        # 【名称】状态 进度 单位 分割线 总数 单位
        _title_info = "[%s] %s " % (self.title, self.status)
        _current_info = "%.2f %s " % (self.transferred/self.chunk_size, self.unit)
        _total_info = "%.2f %s" % (self.toBeTransferred/self.chunk_size, self.unit)

        _info = _title_info + _current_info + self.seq + _total_info
        return _info

    def __get_bar(self):
        # 【名称】状态 进度 百分号 进度符号
        _title_info = "[%s] %s " % (self.title, self.status)
        _rate = "%.2f%s" % (float(self.transferred) / float(self.toBeTransferred) * 100, "%")

        _info = _title_info + _rate + "=" * int(float(self.transferred) / float(self.toBeTransferred) * 50)
        return _info


def seqfy(strs):
    """ 序列化 字符串--->实际效果是，为字符串，添加行号，返回字符串
    Sampe usage:
        strs = ["", None, u"First-line\nSecond-line\nThird-line", u"没有换行符"]
        for s in strs:
            print "---"
            result = seqfy(s)
            print result
            print unseqfy(result)
    """

    if not strs:
        return

    result = ""
    seq = 1
    ss = strs.split("\n")
    for i in ss:
        if i:
            result = "".join([result, str(seq), ".", i, "\n"])
            seq = seq + 1
    return result


def unseqfy(strs):
    """ 反序列化字符串--->实际效果是，去掉每行字符串前面的行号， 返回字符串
    """
    if not strs:
        return

    result = ""
    ss = strs.split("\n")
    for i in ss:
        raw = i.split(".", 1)
        if len(raw) == 2:
            try:
                int(raw[0])
            except:
                result = "".join([result, i, "\n"])
            else:
                result = "".join([result, raw[1], "\n"])
        else:
            result = "".join([result, raw[0], "\n"])

    return result


def stepfy(strs):
    """ 步骤化 字符串 --->实际效果是, 依据 序列化的字符串，转换为 Step_%s_info 的字典， 返回字典
    Sample usage:
        test_strs = [
        "",
        None,
        u"First-line\nSecond-line\nThird-line",
        u'1.First-line\n2.Second-line\n3.Third-line\n',
        u'3.没有换行符',
        u'3.有换行符\n',
        "asdfasdfsdf",
        "1.asdfasdfsdf\n2.sodfi",
        "1.1.dfasdfahttp://192.168.1.1sdfsdf2.1.1.1.1\n",
        "dfasdfahttp://192.168.1.1sdfsdf2.1.1.1.1\n",
        ]
        for i in test_strs:
            steps = stepfy(i)
            un = unstepfy(steps)
            print "string: %r" %i
            print "stepfy: %s" %steps
            print "unstepfy: %r\n" %un
    """

    result = {}
    prog_step = re.compile("^\d+\.")

    if not strs:
        return result

    raws = strs.split("\n")
    for raw in raws:
        step_num = raws.index(raw) + 1
        raw = prog_step.sub("",raw)
        if raw:
            result["Step_%s_info" %step_num] = raw
    return result


def unstepfy(sdict):
    """反步骤化 字符串--->实际效果是, 依据 stepfy返回的字典数据，进行反步骤化，还原数据
    """
    if not sdict:
        return ""

    if not isinstance(sdict, dict):
        return sdict

    result = []
    for k,v in sdict.items():
        num = k.split("_")[1]
        result.append("%s.%s\n" % (num, v))

    if result:
        tmp = CommonUtils.get_sorted_list(result)
        f = lambda x,y: x + y

        return p_compat.reduce(f, tmp)


def map_function(func_str, fw_action_addtion=None, bw_action_addtion=None, alias_func=None):
    """ Sample usage:
        print map_function('set',alias_func = "ini_items");# -> ini_items
        print map_function('set',fw_action_addtion="action_steps_",bw_action_addtion="_for_upd",alias_func = "ini_items"); # -> action_steps_ini_items_for_upd
        print map_function('set(a=1,b=2,c=Test())',"action_steps_","_for_upd","ini_items");# -> action_steps_ini_items_for_upd(a=1,b=2,c=Test())
        print map_function('set("login",a="good",b=Test())',"action_steps_","_for_upd");# -> action_steps_set_for_upd("login",a="good",b=Test())
    """

    split_action_value = re.compile("^(\w+)(\((.*)\)$)?")
    matched   = split_action_value.match(func_str)

    if matched:
        action = matched.group(1).lower()
        value = matched.group(2)
        # params = matched.group(3)

        if alias_func:
            action = alias_func
        if fw_action_addtion:
            action = fw_action_addtion + action
        if fw_action_addtion:
            action = action + bw_action_addtion

        if value:
            return action+value
        else:
            return action



#! python3
# -*- encoding: utf-8 -*-

__title__ = "rtsf"
__short_desc__ = "a test service framework"
__version__ = "3.0.0"
__autor__ = u"RockFeng0"
__author_email__ = "281589463@qq.com"
__license__ = 'MIT'

HOME_PAGE = "https://github.com/RockFeng0/rtsf"



#! python3
# -*- encoding: utf-8 -*-

import json

try:
    FileNotFoundError = FileNotFoundError
except NameError:
    FileNotFoundError = IOError

try:
    JSONDecodeError = json.decoder.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError


class MyBaseError(BaseException):
    pass


class FileFormatError(MyBaseError):
    pass

class ModelFormatError(MyBaseError):
    pass


class ParamsError(MyBaseError):
    pass

class ResponseError(MyBaseError):
    pass


class ParseResponseError(MyBaseError):
    pass


class ValidationError(MyBaseError):
    pass


class InstanceTypeError(MyBaseError):
    pass


class NotFoundError(MyBaseError):
    pass


class DirectoryNotFound(NotFoundError):
    pass


class FunctionNotFound(NotFoundError):
    pass


class VariableNotFound(NotFoundError):
    pass


class ApiNotFound(NotFoundError):
    pass


class SuiteNotFound(NotFoundError):
    pass


class TestcaseNotFound(NotFoundError):
    pass




#! python3
# -*- encoding: utf-8 -*-

import os
import sys
import logging
from rtsf import p_exception

from colorama import Back, Fore, Style, init
from colorlog import ColoredFormatter
init(autoreset=True)


def coloring(msg, color="WHITE"):
    fore_color = getattr(Fore, color.upper())
    return fore_color + msg


def color_print(msg, color="WHITE"):
    fore_color = getattr(Fore, color.upper())
    print(fore_color + msg)


class AppLog(object):
    """ record the logs with your preference  """
    def __init__(self, logger_name=None, log_file=None, fmt=None, log_level=logging.DEBUG, has_color=False):
        self._logger = logging.getLogger(logger_name)
        self.has_color = has_color

        self.fmt = fmt if fmt else u"%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)-8s: %(message)s"
        # self.formatter = logging.Formatter(u'%(asctime)s %(levelname)-8s: %(message)s')
        self._format = logging.Formatter(self.fmt)

        # hide traceback when log level is INFO/WARNING/ERROR/CRITICAL
        if log_level >= logging.INFO:
            sys.tracebacklimit = 0

        # 添加文件日志
        if log_file:
            self.to_file(log_file)

        # 默认打印console日志
        self.to_console()
        self._logger.setLevel(log_level)

    def get_logger(self):
        return self._logger

    # @property
    # def debug(self):
    #     return self._tolog("debug")
    #
    # @property
    # def info(self):
    #     return self._tolog("info")
    #
    # @property
    # def warning(self):
    #     return self._tolog("warning")
    #
    # @property
    # def error(self):
    #     return self._tolog("error")
    #
    # @property
    # def critical(self):
    #     return self._tolog("critical")
    #
    # def _tolog(self, level):
    #     """ log with different level """
    #     def wrapper(msg):
    #         if self.has_color:
    #             color = self.log_colors[level.upper()]
    #             getattr(self._logger, level.lower())(coloring("- {}".format(msg), color))
    #         else:
    #             getattr(self._logger, level.lower())(msg)
    #
    #     return wrapper
               
    def to_file(self, file_path):
        """ 添加文件日志 记录Debug级别日志
        :return:
        """

        # 确保在添加新的文件句柄之前移除旧的文件句柄
        tmp = [hd for hd in self._logger.handlers if not isinstance(hd, logging.FileHandler)]
        self._logger.handlers = tmp

        if os.path.isdir(os.path.abspath(os.path.dirname(file_path))):
            fh = logging.FileHandler(file_path, mode='a')
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(self._format)
            self._logger.addHandler(fh)
        else:
            raise p_exception.DirectoryNotFound(file_path)
    
    def to_console(self):
        """ 添加 控制台 日志， 打印Info级别日志
        :return:
        """

        # 确保在添加新的控制台句柄之前移除旧的控制台句柄
        tmp = [hd for hd in self._logger.handlers if not isinstance(hd, logging.StreamHandler)
               or isinstance(hd, logging.FileHandler)]
        self._logger.handlers = tmp

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # if self.has_color:
        #     self.log_colors = {
        #         'DEBUG':    'cyan',
        #         'INFO':     'green',
        #         'WARNING':  'yellow',
        #         'ERROR':    'red',
        #         'CRITICAL': 'red',
        #     }
        #
        #     color_formatter = ColoredFormatter(u"%(log_color)s%(bg_white)s#" + self.fmt,
        #         datefmt=None,
        #         reset=True,
        #         log_colors=self.log_colors
        #     )
        #
        #     ch.setFormatter(color_formatter)
        # else:
        #     ch.setFormatter(self._format)
        ch.setFormatter(self._format)
        self._logger.addHandler(ch)


if __name__ == '__main__':
    alg = AppLog(logger_name="TestApp1")
    console = alg.get_logger()
    console.debug("console debug")
    console.info("console info")
    console.warning("console warning")
    console.error("console error")

    alg.to_file(file_path=r"D:\auto\buffer\test_sdfsdf_0.log")  # 没有记录
    alg.to_file(file_path=r"D:\auto\buffer\test_sdfsdf_1.log")  # 有记录
    console.debug("console debug - to file")
    console.info("console info - to file")
    console.warning("console warning - to file")
    console.error("console error - to file")

    logger = AppLog(logger_name="TestApp1", log_file=r"D:\auto\buffer\test_sdfsdf_2.log").get_logger()
    logger.debug("logger debug")
    logger.info("logger info")
    logger.warning("logger warning")
    logger.error("logger error")



#! python3
# -*- encoding: utf-8 -*-

import os
import io
import time
import codecs
from jinja2 import Template
from rtsf import p_compat
from rtsf.p_applog import logger
from rtsf.p_common import FileSystemUtils,DateTimeUtils
from rtsf import __about__


class HtmlReporter(object):
    def __init__(self, device_id="", dir_name=''):
        
        result_name = "report" if not device_id else "report_{}".format(device_id)
        result_path = dir_name if os.path.isdir(dir_name) else os.getcwd()
        
        self.result_path = os.path.join(result_path, result_name)
        self.case_log_path = os.path.join(self.result_path,"caselogs")
#         self.screen_shot_path = os.path.join(self.result_path,"screenshots")  
        self.summary = []      
                    
    def start_test(self, module_name, case_name, resp_tester, tester):
        """
        @param module_name: test set name or test module name
        @param case_name: normal text case name, may contain '@#$~%^(+=-)&* ...'
        @param resp_tester: name of responsible tester
        @param tester: name of tester who run this case  
        """
        self.meta_data = {
            "module_name": module_name,
            "raw_case_name": case_name,
            "case_name": FileSystemUtils.get_legal_filename(case_name),
            "status": "pass",
            "resp_tester": resp_tester,
            "tester": tester,
            "start_at": time.time(),
            "end_at": "",
            }
        
        FileSystemUtils.mkdirs(self.case_log_path)
#         FileSystemUtils.mkdirs(self.screen_shot_path)
        
        log_file = self.__get_log_file()
        with codecs.open(log_file, "a", "utf-8") as f:
            f.write(u"\n**************  %s [%s]  ***************\n" %(u"Case Log From Rock4 Test Service Framework",self.meta_data['case_name']))
    
    def stop_test(self):
        self.meta_data["end_at"] = time.time()
        HtmlReporter.add_report_data(list_all=self.summary, **self.meta_data)
            
    def step_info(self, info, msg):
        
        if isinstance(msg, p_compat.bytes):            
            try:
                unicode_msg = msg.decode('utf-8')                
            except:
                raise Exception("Log message not unicode or utf-8.")
        else:
            unicode_msg = msg
        
        info = info.upper()
        log_file = self.__get_log_file()
        with codecs.open(log_file, "a", "utf-8") as f:
            if info == "SECTION":      
                f.write(u"\n%-20s\t%-10s\t%s\n" %(DateTimeUtils.get_stamp_datetime_coherent(),info, unicode_msg))
            elif info in ["NORMAL","STEP","PASS"]:
                f.write(u"%-20s\t%-10s\t%s\n" %(DateTimeUtils.get_stamp_datetime_coherent(),info,unicode_msg))
            elif info in ["ERROR","FAIL"]:
                f.write(u"%-20s\t%-10s\t%s\n" %(DateTimeUtils.get_stamp_datetime_coherent(),info,unicode_msg))
                self.meta_data["status"] = "Fail"

    def __get_log_file(self):
        log_file_name = u"%s_%s.log" %(self.meta_data['case_name'], DateTimeUtils.get_stamp_date())
        log_file = os.path.join(self.case_log_path,log_file_name)
        if not os.path.isfile(log_file):
            FileSystemUtils.mkdirs(self.case_log_path)
        return log_file 

    def generate_html_report(self, proj_name, proj_module = None):
        html_results = []               
        all_summary = HtmlReporter.get_summary(self.summary, proj_name = proj_name)
        
        for summary in all_summary:
            html_report = os.path.join(self.result_path, u"[{}]{}_{}.html".format(FileSystemUtils.get_legal_filename(summary["project_name"]),
                                                                                    FileSystemUtils.get_legal_filename(summary["module_name"]), 
                                                                                DateTimeUtils.get_stamp_datetime_coherent(),
                                                                                ))        
            if proj_module == None:                
                html_results.append(HtmlReporter.render_html(html_report, summary))
                
            elif summary["module_name"] == proj_module:
                html_results.append(HtmlReporter.render_html(html_report, summary))
                break
            else:
                summary = {}
        return html_results
    
    @staticmethod
    def render_html(report_file_path, summary):        
        html_report_template = os.path.join(os.path.abspath(os.path.dirname(__file__)),"templates","default_report_template.html")
        
        with io.open(html_report_template, "r", encoding='utf-8') as f_r:
            template_content = f_r.read()            
            with io.open(report_file_path, 'w', encoding='utf-8') as f_w:
                rendered_content = Template(template_content).render(summary)            
                f_w.write(rendered_content)
        
#         with open(os.path.join(os.path.dirname(report_file_path),'result.json'), 'w') as f:
#             f.write(str(summary.get("dict_report","")))
        
        return report_file_path
    
    @staticmethod
    def get_summary(list_all=[], **kwargs):
        """ summarize the report data
            @param list_all: a list which save the report data
            @param kwargs: such as
                show_all:    True/False   report show all status cases
                proj_name:   project name 
                home_page:   home page url
        
        """
        all_summary = []
               
        for module in list_all:
            summary = {
                        "module_name": module['Name'],
                        "show_all": kwargs.get("show_all",True),
                        "project_name": kwargs.get("proj_name","TestProject"),
                        "home_page": kwargs.get("home_page",__about__.HOME_PAGE),
                        "start_time": "",
                        "end_time": "",
                        "duration_seconds": "",
                        "total_case_num": len(module["TestCases"]),
                        "pass_cases_num": 0,
                        "fail_cases_num": 0,
                        "details": []
                    }
                         
            for case in module["TestCases"]:
                case_detail = {}
                case_detail["linkurl"] = "./caselogs/%s_%s.log" %(case["case_name"],case["exec_date"])
                
                if case["status"].lower() == "pass":
                    summary["pass_cases_num"] += 1
                    case_detail["c_style"] = "tr_pass"
                else:
                    summary["fail_cases_num"] += 1
                    case_detail["c_style"] = "tr_fail"
                
                case_detail.update(case)
            
                summary["details"].append(case_detail)                       
             
            try:
                st = module["TestCases"][0].get("start_at")
                et = module["TestCases"][-1].get("end_at")
                
                summary["start_time"] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(st))
                summary["end_time"] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(et))
                summary["duration_seconds"] = float("%.2f" % (et - st))
            except Exception as _:
                logger.log_warning("Will set 'start_at' and 'end_at' to 'None'")
                (summary["start_time"], summary["end_time"], summary["duration_seconds"]) = (None, None, None)
                    
            if summary["fail_cases_num"] > 0:
                summary["dict_report"] = {
                    "result": 0,
                    "message": "failure",
                    "pass": summary["pass_cases_num"],
                    "fail": summary["fail_cases_num"]
                }
            else:
                summary["dict_report"] = {
                    "result": 1,
                    "message": "success",
                    "pass": summary["pass_cases_num"],
                    "fail": summary["fail_cases_num"]
                }
             
            all_summary.append(summary)
            
        return all_summary
            
    @staticmethod
    def add_report_data(list_all=[], module_name="TestModule", **kwargs):
        """ add report data to a list
            @param list_all: a list which save the report data
            @param module_name: test set name or test module name
            @param kwargs: such as
                case_name:   testcase name
                status:      test result, Pass or Fail
                resp_tester: responsible tester who write this case
                tester:      tester who execute the test
                start_at:    tester run this case at time 
                end_at:      tester stop this case at time
        """
        start_at = kwargs.get("start_at")        
        case_name = kwargs.get("case_name", "TestCase")
        raw_case_name = kwargs.get("raw_case_name", "TestCase")
                
        exec_date_time = time.localtime(start_at)
        execdate = time.strftime("%Y-%m-%d", exec_date_time)
        exectime = time.strftime("%H:%M:%S", exec_date_time)
        
        _case_report = {
                'resp_tester': kwargs.get("resp_tester", "administrator"),
                'tester': kwargs.get("tester", "administrator"),
                'case_name': case_name,
                'raw_case_name': raw_case_name,
                'status': kwargs.get("status", "Pass"),
                'exec_date': execdate,
                'exec_time': exectime,
                'start_at': start_at,
                'end_at': kwargs.get("end_at"),
            }
                
        for module in list_all:
            if module_name != module["Name"]:
                continue
            
            for case in module["TestCases"]:
                if raw_case_name == case["raw_case_name"]:
                    case.update(_case_report)
                    return list_all
            
            module["TestCases"].append(_case_report)
            return list_all
        
        list_all.append({"Name": module_name, "TestCases": [_case_report]})
        return list_all



#! python3
# -*- encoding: utf-8 -*-


from rtsf.p_report import HtmlReporter
from rtsf.p_applog import AppLog


class Tracer(HtmlReporter, AppLog):
    def __init__(self, **kwargs):
        self.__clear = False
        HtmlReporter.__init__(self, device_id=kwargs.get('device_id', ""), dir_name=kwargs.get('dir_name', ""))
        
        AppLog.__init__(self, logger_name=kwargs.get('logger_name'))
    
    def start(self,module_name, case_name, resp_tester, tester):
        if self.__clear:
            return
        self.start_test(module_name, case_name, resp_tester, tester)        
        self.log_info(u"-------\n\t#### Starting test {}: {} {} {}".format(module_name, case_name, resp_tester, tester))

    def section(self, strs):
        if self.__clear:
            return        
        self.step_info("section", self.__deal_str(strs))
        #self.log_info(self.__deal_str(strs))

    def normal(self, strs):
        if self.__clear:
            return        
        self.step_info("normal", self.__deal_str(strs))
        self.log_info(self.__deal_str(strs))

    def step(self, strs):
        if self.__clear:
            return
        self.step_info("step", self.__deal_str(strs))
        self.log_info(self.__deal_str(strs))

    def ok(self, strs):
        if self.__clear:
            return
        self.step_info("pass", self.__deal_str(strs))
        self.log_info(self.__deal_str(strs))

    def fail(self, strs):
        if self.__clear:
            return
        self.step_info("fail", self.__deal_str(strs))
        self.log_info(self.__deal_str(strs))

    def error(self, strs):
        if self.__clear:
            return
        self.step_info("error", self.__deal_str(strs))
        self.log_error(self.__deal_str(strs))

    def stop(self):
        if self.__clear:
            return
        self.stop_test()
        self.log_info(u"\n\t## Stopped test")
    
    def _switch_off(self):
        self.__clear = True
        
    def _switch_on(self):
        self.__clear = False
            
    def __deal_str(self,strs):
        if isinstance(strs, str):
            try:
                return strs.decode("utf-8")
            except:
                pass
        return strs    


tracer = Tracer()



#! python3
# -*- encoding: utf-8 -*-

import os
import sys
import unittest
import threading
import multiprocessing
from functools import partial
from rtsf.p_applog import logger
from rtsf.p_tracer import Tracer
from rtsf.p_testcase import YamlCaseLoader, parse_project_data
from rtsf import p_testcase, p_compat, p_exception


class TestCase(unittest.TestCase):
    """ create a testcase.
    """
    def __init__(self, test_runner, testcase_dict, variables):
        super(TestCase, self).__init__()
        self.test_runner = test_runner
        self.testcase_dict = testcase_dict.copy()
        self.variables = variables

    def runTest(self):
        """ run testcase and check result.
        """
        self.test_runner._run_test(self.testcase_dict, self.variables)


class TestSuite(unittest.TestSuite):
    """ create test suite with a testset, it may include one or several testcases.
        each suite should initialize a separate Runner() with testset config.
    @param
        (dict) testset
            {
                "name": "testset description",
                "project": {
                    "name": "project name",
                    "module": "testset description"
                    "data":[
                                {'csv': 'username_password.csv', 'by': 'Sequential'}, 
                                {'csv': 'devices.csv', 'by': 'Sequential'}
                            ]
                },
                "cases": [
                    {
                        "name": "testcase description",
                        "tester": "",    # optional
                        "responsible": "",    # optional
                        "pre_command": [],    # optional
                        "steps": [],      
                        "post_command": {},     # optional
                        "verify": []         # optional
                    },
                    testcase12
                ]
            }
    """            
    def __init__(self, testset, runner_cls):
        super(TestSuite, self).__init__()
         
        file_path    = testset.get("file_path")
        project      = testset.get("project")
        testcases    = testset.get("cases", [])        
        project_data = project.pop("data", [])
        
        test_runner = self.test_runner = runner_cls()
        if not isinstance(test_runner._default_devices, (list, tuple)):            
            raise TypeError("_default_devices not a list or tuple.")

        test_runner.init_runner(parser=p_testcase.TestCaseParser(file_path=file_path),
                            tracers={device: Tracer(device_id=device, dir_name=os.path.dirname(os.path.abspath(file_path))) for device in test_runner._default_devices},
                            projinfo=project
                            )
        
        for data_variables_dict in parse_project_data(project_data, file_path) or [{}]:
            for testcase_dict in testcases:                        
                self._add_test_to_suite(testcase_dict["name"], test_runner, testcase_dict, data_variables_dict)
                             
    def _add_test_to_suite(self, testcase_name, test_runner, testcase_dict, variables):
        if p_compat.is_py3:
            TestCase.runTest.__doc__ = testcase_name
        else:
            TestCase.runTest.__func__.__doc__ = testcase_name
        
        test = TestCase(test_runner, testcase_dict, variables)
        [self.addTest(test) for _ in range(int(testcase_dict.get("times", 1)))]        
        
    @property
    def tests(self):                
        return self._tests


class TaskSuite(unittest.TestSuite):
    """ create task suite with specified testcase path.
        each task suite may include one or several test suite.
    """
    def __init__(self, testsets, runner_cls):
        """
        @params
            testsets (dict/list): testset or list of testset
                testset_dict
                or
                [
                    testset_dict_1,
                    testset_dict_2,
                    {
                        "name": "desc1",
                        "config": {},
                        "api": {},
                        "testcases": [testcase11, testcase12]
                    }
                ]
            mapping (dict):
                passed in variables mapping, it will override variables in config block
        """
        super(TaskSuite, self).__init__()

        if not testsets:
            raise p_exception.TestcaseNotFound

        if isinstance(testsets, dict):
            testsets = [testsets]
        
        self.suite_list = []
        for testset in testsets:
            suite = TestSuite(testset, runner_cls)
            self.addTest(suite)
            self.suite_list.append(suite)

    @property
    def tasks(self):
        return self.suite_list


def init_test_suite(path_or_testsets, runner_cls):
    if not p_testcase.is_testsets(path_or_testsets):
        YamlCaseLoader.load_dependencies(path_or_testsets)        
        testsets = YamlCaseLoader.load_files(path_or_testsets)
    else:
        testsets = path_or_testsets

    return TaskSuite(testsets, runner_cls)


class TestRunner(object):

    def __init__(self, **kwargs):
        """ initialize test runner
        @param (dict) kwargs: key-value arguments used to initialize TextTestRunner            
        """
        runner_cls = kwargs.pop("runner", Runner)
        
        if not callable(runner_cls) and not isinstance(runner_cls(), Runner):
            raise p_exception.InstanceTypeError("Invalid runner, must be instance of Runner.")
        
        self._runner_cls = runner_cls
        self.runner = unittest.TextTestRunner(**kwargs)

    def run(self, path_or_testsets):
        """ start to run test with varaibles mapping
        @param path_or_testsets: YAML/JSON testset file path or testset list
            path: path could be in several type
                - absolute/relative file path
                - absolute/relative folder path
                - list/set container with file(s) and/or folder(s)
            testsets: testset or list of testset
                - (dict) testset_dict
                - (list) list of testset_dict
                    [
                        testset_dict_1,
                        testset_dict_2
                    ]
        """
                
        try:
            self._task_suite =init_test_suite(path_or_testsets, self._runner_cls)
        except p_exception.TestcaseNotFound:
            logger.log_error("Testcases not found in {}".format(path_or_testsets))
            sys.exit(1)

        self.text_test_result = self.runner.run(self._task_suite)        
        return self
    
    def gen_html_report(self):
        html_report = []
        for suite in self._task_suite.tasks:
            proj_name = suite.test_runner.proj_info["name"]
            reporters = suite.test_runner.tracers.values()
            
            for reporter in reporters:
                html_report.extend(reporter.generate_html_report(proj_name, proj_module=None))
        return html_report        


class Runner(object):
    
    def __init__(self):
        """
        @note: maybe override variables
            _default_devices -> list type; to genrate tracer map, format is `{device_id: tracer_obj}`
                            e.g.
                                 default {"":tracer_obj} use to generate report for local host; 
                                 {"192.168.0.1:5555":tracer_obj1, "192.168.0.2:5555":tracer_obj2} use to generate report for remote host if run with mutilple process                                
            _default_drivers -> list type; to define driver map, format is `(device_id: driver)`
                            e.g.
                                default ("", None) use to run case with a driver;
                                [("192.168.0.1:5555":selenium_driver), ("192.168.0.2:5555":appium_driver), ...] use for multiple process to run case with specified drivers                      
        """
        self._default_devices = [""]
        self._default_drivers = [("", None)]
        self._local_driver = True
    
    def init_runner(self, parser, tracers, projinfo):
        """ initial some instances for preparing to run test case
        @note:  should not override
        @param parser: instance of TestCaseParser
        @param tracers: dict type for the instance of Tracer. Such as {"":tracer_obj} or {"192.168.0.1:5555":tracer_obj1, "192.168.0.2:5555":tracer_obj2} 
        @param proj_info: dict type of test case.  use like:  self.proj_info["module"], self.proj_info["name"]
            yaml case like: 
                - project:
                    name: xxx
                    module: xxxx
            dict case like:
                {"project": {"name": xxx, "module": xxxx}}            
                
        """
        self.parser = parser
        self.tracers = tracers
        self.proj_info = projinfo
        
    def run_test(self, testcase_dict, variables, driver_map):
        """ define how to run a case. override this method
        @param testcase_dice:  yaml case
        @param driver_map:  device id map to a driver 
              
        """
        fn, _ = driver_map
        reporter = self.tracers[fn]
        
        parser = self.parser
        parser.update_binded_variables(variables)
        
        case_name = parser.eval_content_with_bind_actions(testcase_dict.get("name", u'rtsf'))
        reporter.start(self.proj_info["module"], case_name, testcase_dict.get("responsible", u"rock feng"), testcase_dict.get("tester", u"rock feng"))
        reporter.log_debug(u"===== run_test\n\t{}".format(testcase_dict))
        
        reporter.section(u"------------section ok")
        reporter.step(u"step ok")
        reporter.normal(u"normal ok")
        reporter.stop()
        
        return reporter
    
    def _run_test(self, testcase_dict, variables={}):
        """ guide the running case
        @param testcase_dice:  yaml case
        @param variables: dict type; this is defined the variables for the data-driven test
                            e.g.
                                default {} use to run case without data-driven
                                {"username":"test1","password":"123456"}
        """
        if self._local_driver:
            self.run_test(testcase_dict, variables, self._default_drivers[0])
        else:
            self._drivers = []
            self._run_grid_multithread(partial(self.run_test, testcase_dict, variables), self._default_drivers)
            
    def _run_grid_multiprocess(self, func, iterables):
        """ running case with mutil process to support selenium grid-mode(multiple web) and appium grid-mode(multiple devices). 
        @param func:  function object
        @param iterables:  iterable objects
        """
        multiprocessing.freeze_support()
        pool = multiprocessing.Pool()        
        pool_tracers = pool.map(func, iterables)
        pool.close()
        pool.join()
        
        # 传递给 pool.map的 实例对象，内存地址发生变化， 因此，这里在运行结束后，重新定义 self.tracers 
        self.tracers = dict(zip(self._default_devices, pool_tracers))
    
    def _run_grid_multithread(self, func, iterables):
        """ running case with mutil process to support selenium grid-mode(multiple web) and appium grid-mode(multiple devices). 
        @param func:  function object
        @param iterables:  iterable objects
        """
        
        f = lambda x: threading.Thread(target=func, args=(x,))
        threads = map(f, iterables)
        for thread in threads:
            thread.setDaemon(True)
            thread.start()
            thread.join()



#! python3
# -*- encoding: utf-8 -*-

import os
import io
import sys
from shutil import rmtree
from rtsf import __about__
from setuptools import find_packages, setup, Command

here = os.path.abspath(os.path.dirname(__file__))

with io.open("README.md", encoding='utf-8') as f:
    long_description = f.read()
    
install_requires = [
    "PyYAML",
    "Jinja2",
    "colorama",
    "colorlog",
    "urllib3",
]


class UploadCommand(Command):
    """ Build and publish this package.
        Support setup.py upload. Copied from requests_html.
    """

    user_options = []

    @staticmethod
    def status(s):
        """Prints things in green color."""
        print("\033[0;32m{0}\033[0m".format(s))

    def initialize_options(self):
        """ override
        """
        pass

    def finalize_options(self):
        """ override
        """
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Publishing git tags…')
        os.system('git tag v{0}'.format(__about__.__version__))
        os.system('git push --tags')

        sys.exit()
        
setup(
    name=__about__.__title__,
    version=__about__.__version__,
    description=__about__.__short_desc__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=__about__.__autor__,
    author_email=__about__.__author_email__,
    url=__about__.HOME_PAGE,
    license=__about__.__license__,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    packages=find_packages(exclude=()),
    package_data={
        'rtsf': ["templates/*"],
    },
    keywords='test requests locust HTTP api selenium appium uiautomation',
    install_requires=install_requires,
    extras_require={},
    # $ setup.py upload support.
    cmdclass={
        'upload': UploadCommand
    }
)





#! python3
# -*- encoding: utf-8 -*-

import sys

# -------
# Pythons
# -------

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

try:
    import simplejson as json
except ImportError:
    import json

# ---------
# Specifics
# ---------

if is_py2:
    from urllib3.packages.ordered_dict import OrderedDict
    import ConfigParser
    reduce = reduce
    
    builtin_str = str
    bytes = str
    str = unicode      
    basestring = basestring
    numeric_types = (int, long, float)
    integer_types = (int, long)
    
    xrange = xrange

elif is_py3:
    
    from collections import OrderedDict
    import configparser as ConfigParser
    from functools import reduce
    
    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    numeric_types = (int, float)
    integer_types = (int,)
    
    xrange = range
